#include <Eigen/Eigen>
#include <unsupported/Eigen/SparseExtra>
#include <fstream>
#include <glog/logging.h>
#include <cmath>
#include <string>
#include <sstream>
#include <vector>

#include "util/util.h"
#include "util/read_write_tensor.h"
#include "util/timer.h"
#include "util/stringprintf.h"

void ReadTensorSlice(std::ifstream& infile, Adv_Tensor_Type& slice) {
  if (!infile.is_open())
    LOG(FATAL) << "Cannot read in stream: ";
  int tensor_dim;
  infile.read(reinterpret_cast<char*>(&tensor_dim), sizeof(int));
  uint64_t nnzs, outS,innS;
  infile.read(reinterpret_cast<char*>(&nnzs), sizeof(uint64_t));
  infile.read(reinterpret_cast<char*>(&outS), sizeof(uint64_t));
  infile.read(reinterpret_cast<char*>(&innS), sizeof(uint64_t));
  slice.resize(tensor_dim, tensor_dim);
  slice.makeCompressed();
  slice.resizeNonZeros(nnzs);
  infile.read(reinterpret_cast<char*>(slice.valuePtr()), sizeof(double)*nnzs);
  infile.read(reinterpret_cast<char*>(slice.outerIndexPtr()), sizeof(int)*outS);
  infile.read(reinterpret_cast<char*>(slice.innerIndexPtr()), sizeof(int)*nnzs);
  slice.finalize();
}

void ReadTensor(const std::string file, const int expected_dim ,const int expected_type, 
                   std::vector<Adv_Tensor_Type>* Adv_tensor) {
  std::ifstream infile(file.c_str());
  if (!infile.is_open()) {
    LOG(FATAL) << "Cannot open tensor file: " << file;
  }
  ReadTensor(infile, expected_dim, expected_type, Adv_tensor);
  LOG(INFO) << "Read completed..";
  infile.close();
}

void WriteTensorSlice(const Adv_Tensor_Type& slice, std::ofstream& out) {
  if (!out.is_open()) {
    LOG(FATAL) << "Cannot write to stream:";
  }
  const int tensor_dim = slice.rows(); 
  // first write out the tensor dimention.
  out.write(reinterpret_cast<const char *>(&tensor_dim), sizeof(int));
  uint64_t nnzs = slice.nonZeros();
  uint64_t outS = slice.outerSize();
  uint64_t innS = slice.innerSize();
  out.write(reinterpret_cast<const char *>(&nnzs), sizeof(uint64_t));
  out.write(reinterpret_cast<const char *>(&outS), sizeof(uint64_t));
  out.write(reinterpret_cast<const char *>(&innS), sizeof(uint64_t));
  // Write the matrix values.
  out.write(reinterpret_cast<const char *>(slice.valuePtr()), sizeof(double)*nnzs);
  // Output the index values, the default index value type for eigen is int.
  out.write(reinterpret_cast<const char *>(slice.outerIndexPtr()), sizeof(int)*outS);
  out.write(reinterpret_cast<const char *>(slice.innerIndexPtr()), sizeof(int)*nnzs);
}

void WriteTensor(const std::vector<Adv_Tensor_Type>& Adv_tensor,
                 const int tensor_type, const std::string file) {
  
  std::ofstream out(file);
  if (!out.is_open()) {
    LOG(FATAL) << "Cannot write to file: " << file;
  }
  WriteTensor(Adv_tensor, tensor_type, out);
  out.close();
}

uint64_t countTensorBytes(const std::vector<Adv_Tensor_Type>& Adv_tensor) {
  uint64_t result = 0;

  int tensor_dim = Adv_tensor.size();
  result += sizeof(int)*2;

  //out.write(reinterpret_cast<const char *>(&tensor_dim), sizeof(int));
  //out.write(reinterpret_cast<const char *>(&type_), sizeof(int));
  // Write each sparse matrix.
  for (int i = 0; i < tensor_dim; i++) {
    uint64_t nnzs = Adv_tensor[i].nonZeros();
    uint64_t outS = Adv_tensor[i].outerSize();
    uint64_t innS = Adv_tensor[i].innerSize();
    result += sizeof(uint64_t)*3;

    //out.write(reinterpret_cast<const char *>(&nnzs), sizeof(uint64_t));
    //out.write(reinterpret_cast<const char *>(&outS), sizeof(uint64_t));
    //out.write(reinterpret_cast<const char *>(&innS), sizeof(uint64_t));
    result += sizeof(double)*nnzs;

    //out.write(reinterpret_cast<const char *>(Adv_tensor[i].valuePtr()), sizeof(double)*nnzs);
    result += sizeof(int)*outS;
    result += sizeof(int)*nnzs;
    //out.write(reinterpret_cast<const char *>(Adv_tensor[i].outerIndexPtr()), sizeof(int)*outS);
    //out.write(reinterpret_cast<const char *>(Adv_tensor[i].innerIndexPtr()), sizeof(int)*nnzs);
  }
  
  return result;
}

void ReadTensorFromMatlab(const std::string& file_name, const int basis_dim, 
std::vector<Adv_Tensor_Type>* new_tensor) {
  
  new_tensor->reserve(basis_dim);
  for (int k = 0; k < basis_dim; k++) {
    Adv_Tensor_Type Ck(basis_dim, basis_dim);
    Ck.setZero();
    new_tensor->emplace_back(Ck);
  }
  
  const int N = basis_dim*basis_dim*basis_dim;
  double *U = (double*)malloc(N * sizeof(double));
  std::ifstream infile;
  infile.open(file_name.c_str(), std::ios::in | std::ios::binary);
  infile.read((char*)U, N*sizeof(double));
  infile.close();
  for (int k = 0; k < basis_dim; k++) {
    for (int j = 0; j < basis_dim; j++) {
      for (int i = 0; i < basis_dim; i++) {
        int index = k * basis_dim * basis_dim + j * basis_dim + i;
        if (std::abs(U[index]) > 1e-10) {
          SetMatrixEntry(&(*new_tensor)[k], i, j, U[index]);
        }
      }
    }
  }
  delete [] U;
  infile.close();
}
